/*
 * mdb.h for mdb.  This must be included first, if used.
 */ 

/* Gather kernel headers from all over. The <kernel> directory must be in
 * cc's search path.
 */
#include <minix/config.h>
#include <minix/const.h>
#include <sys/types.h>
#include <minix/type.h>
#include <limits.h>
#include <errno.h>
#include "kernel/const.h"
#include "kernel/type.h"
#include "kernel/proc.h"
#undef printf			/* defined as printk in <kernel/const.h> */

#include <sys/ptrace.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <stdio.h>

/* Support BSD Symbols for gcc. */
/* Hack around different naming conventions */
#ifdef GNU_SYMBOL
#include "a.out.gnu.h"
#include "a.out.minix.h"
#define N_NAME n_un.n_name
#define EXEC_T aexec
#define GET_CLASS(a) (a.n_type   & N_TYPE)
#else
#include <a.out.h>
#define N_NAME n_name
#define EXEC_T exec
#define GET_CLASS(a) (a.n_sclass & N_SECT)
#endif

/* Attempt to provide better identifiers than CHIP, MACHINE and INTEL_32BITS
 * to base decisions on.
 */

#if (CHIP == M68000)
#       ifndef __mc68000__
#	define __mc68000__	/* controls processor-dependent stuff */
#	endif
#	if (MACHINE == ATARI)
#		define MINIX_ST	/* controls system-dependent stuff */
#	else
#		error "only the MINIX_ST 1.5.x implementation works on 68K's"
#	endif
#endif

#if (CHIP == INTEL)
#	define __i8086__
#	if INTEL_32BITS
#		define __i80386__
#	endif
#	if (MACHINE == IBM_PC)
#		define MINIX_PC
#	else
#		error "only the MINIX_PC 1.5.x implementation works on *86's"
#	endif
#endif

#define ADDRSIZE (sizeof(PC_MEMBER(prc)))
#define BITSIZE(size) (8 * (size))
#define INTSIZE (sizeof(int))	/* not quite right for cross-debugger */
#define LONGSIZE (sizeof(long))

#ifdef __mc68000__
#	define BYTES_SWAPPED
#endif

#ifdef BYTES_SWAPPED		/* this assumes words are swapped too */
#	define MASK(size) \
		((size) >= LONGSIZE ? -1L : (1L << BITSIZE(size)) - 1L)
#	define SHIFT(size) (int) (BITSIZE(LONGSIZE - (size)))
#else
#	define MASK(size) \
		((size) >= LONGSIZE ? -1L : (1L << BITSIZE(size)) - 1L)
#	define SHIFT(size) (0)
#endif

#ifdef MINIX_ST	
#	define BP_MEMBER(rp) ((rp)->p_reg.fp)
#	define PC_MEMBER(rp) ((rp)->p_reg.pc)
#	define SP_MEMBER(rp) ((rp)->p_reg.sp)
#	define PSW_MEMBER(rp) ((rp)->p_reg.psw)
#ifndef __GNUC__
#	define PC_OFF ((long)&((struct proc *)0)->p_reg.pc)
#	define BP_OFF ((long)&((struct proc *)0)->p_reg.fp)
#	define PSW_OFF ((long)&((struct proc *)0)->p_reg.psw)
#else
#include <stddef.h>
#	define PC_OFF ((long) offsetof(struct proc,p_reg.pc))
#	define BP_OFF ((long) offsetof(struct proc, p_reg.fp))
#	define PSW_OFF ((long) offsetof(struct proc, p_reg.psw))
#endif
#endif

#ifdef MINIX_PC
#	define BP_OFF ((long)&((struct proc *)0)->p_reg.bp)
#	define BP_MEMBER(rp) ((rp)->p_reg.bp)
#	define PC_MEMBER(rp) ((rp)->p_reg.pc)
#	define PC_OFF ((long)&((struct proc *)0)->p_reg.pc)
#endif

#ifdef __i8086__		
#	ifdef __i80386__
#	define N_REG16	4	/* 16 bit registers at start of stackframe */ 
#	else 
#	define N_REG16	2
#	endif
#endif

/* SIzes may be recorded in units other than CLICKs. */
#define msize_to_physb(n) ((phys_bytes) (n) << CLICK_SHIFT)
#define physb_to_msize(n) ((vir_clicks) ((n) >> CLICK_SHIFT))

#if defined(MINIX_PC) && defined(__i80386__)
/* Bruce's kernel now has sizes in bytes.
 * It would be better if the conversion was defined in kernel/const.h or
 * minix/const.h.
 */
#define ksize_to_physb(n) ((phys_bytes) (n))
#else
#define ksize_to_physb(n) ((phys_bytes) (n) << CLICK_SHIFT)
#endif

#ifdef MAIN			/* mdb.c cannot just redefine EXTERN */
#undef EXTERN			/* because that would apply to junk in user.h*/
#define EXTERN
#endif 

#include <setjmp.h>

EXTERN struct proc *prc;
EXTERN long lbuf[1 + sizeof(struct proc)/sizeof(long)], 
	saddr,
	daddr,
	taddr,
	eaddr;
EXTERN int curpid, 
	cursig,
	seg;
#ifdef DEBUG
EXTERN int debug;
#endif 
EXTERN char *prog, 
	*cmd, 
	*cmdstart,
	*core;
EXTERN jmp_buf mainlp;

/* Prototypes created by mkptypes. */
#if defined(__STDC__) || defined(__cplusplus)
# define P_(s) s
#else
# define P_(s) ()
#endif

/* core.c */
int checkobj P_((struct EXEC_T *header));
void prtmap P_((void));
int core_init P_((char *corefn, char *progfn, int rw, int flat));
long myptrace P_((int req, int pid, long addr, long data));

/* mdb.c */
int disp_regs P_((void));
char *skip P_((char *s));
int main P_((int argc, char *argv[]));

/* dis68.c, mdbdis86.c */
long dasm P_((long addr, int count, int symflg));

/* mdbexp86.c */
char *getexp P_((char *buf, long *exp_p, int *seg_p));
long get_reg P_((int pid, long k));
long reg_addr P_((char *s));

/* sym68.c, sym86.c */
void getsyms P_((char *filename));
void listsym P_((char *local_cmd));
void symbolic P_((unsigned long value, int separator));
