/*
 * mdbexp.c - MINIX expresion parser for 8086-386 + M68000
 *
 * Original version by Bruce D. Szablak
 *
 * This free software is provided for non-commerical use. No warrantee
 * of fitness for any use is implied. You get what you pay for. Anyone
 * may make modifications and distribute them, but please keep this header
 * in the distribution. 
 */

#include "mdb.h"
#include <ctype.h>

#if	(CHIP == INTEL)
/* XXX - this file still uses CHIP instead of __XXX_CPU__. */
/* Hack - define some things directly to avoid #include "db/const.h". */
#define CSEG		0x2E	/* 8088 through 80386 */
#define DSEG		0x3E
#endif

FORWARD long value P_((char *s, char **s_p, int *seg_p));
FORWARD long lookup P_((char *s, char **s_p, int *seg_p));

#if (CHIP == INTEL)
# ifdef __i80386__
PRIVATE char regs[] = "gs fs es ds di si bp .. bx dx cx ax .. ip cs ps sp ss";
# else
PRIVATE char regs[] = "es ds di si bp .. bx dx cx ax .. ip cs ps sp ss";
# endif 
#endif

#define idchar(c) (isalpha(c) || isdigit(c) || (c) == '_')
 
PUBLIC char *getexp(buf, exp_p, seg_p)
char *buf;
int *seg_p;
long *exp_p;
{
	long v = 0L;
 
	buf = skip(buf);
	if ((isalpha(*buf) && (isspace(buf[1]) || buf[1]==';'))
		|| *buf == '\n' 
		|| *buf == ';' 
		|| *buf == '/' 
		|| *buf == '!' 
		|| *buf == '?'
		|| *buf == '#')
	{
		*exp_p = 0L;
		return buf;
	}
	v = value(buf, &buf, seg_p);
	buf = skip(buf);
	if (*buf == '+')
	{
		v += value(skip(buf+1), &buf, seg_p);
	}
	else if (*buf == '-')
	{
		v -= value(skip(buf+1), &buf, seg_p);
	}
	*exp_p = v;
	return skip(buf);
}

#if (CHIP == INTEL)
/* Get register for pid at offset k */
PUBLIC long get_reg(pid, k)
int pid;
long k;
{
  long off;
  long val;
  int reg_size; 

  /* calculate size of register */	
  reg_size = ( k < N_REG16 * 2 ) ? 2 : sizeof(reg_t);

  /* adjust offset */
  off = k - (k & (sizeof(long) -1) ); 

  val = myptrace(T_GETUSER, pid, off, 0L);

  if (k & (sizeof(long) -1) ) 
	val >>= BITSIZE(reg_size);
  else
	val &= MASK(reg_size);
  return val;
}


PUBLIC long reg_addr(s)
char *s;
{
	long val;
	char *t;
	char *send;
	char q[3];	

	q[0] = tolower(*s);
	q[1] = tolower(*++s);
	q[2] = '\0';

	t = regs;
	send = regs + sizeof(regs);
	while(t < send) {
	    if ( strncmp(q,t,2) == 0) {
		val = (long)(t - regs); 
		val /= 3L;
		if ( val < N_REG16 -1) 
			val = val * 2;
		else
			val = (N_REG16 -1) * 2 + 
			      (val - N_REG16 +1) * sizeof(reg_t);
		return val;
	    }
	t += 3;
	}

	printf("Unknown register: %s\n", q);
	longjmp(mainlp, 1);
	return 0;		/* NOTREACHED */
}
#else
/* Get register for pid at offset k */
PUBLIC long get_reg(pid, k)
int pid;
long k;
{
  long off;
  long val;
  int reg_size; 

  /* calculate size of register */
  reg_size = 32;

  /* adjust offset */
  off = k - (k & (sizeof(long) -1) ); 

  val = myptrace(T_GETUSER, pid, off, 0L);

  if (k & (sizeof(long) -1) ) 
	val >>= BITSIZE(reg_size);
  else
	val &= MASK(reg_size);
  return val;
}

PUBLIC long reg_addr(s)
	char *s;
{
	long val;

	switch (*s++)
	{
	case	'a': case 'A': val = 32; break;
	case	'd': case 'D': val = 0; break;
	case	'p': case 'P':
		switch (*s)
		{
		case 'c': case 'C': return(PC_OFF);
		case 's': case 'S': return(PSW_OFF);
		/* fallthrough */
		}
	default: s--; goto error;
	}
	if (*s >= '0' && *s <= '7')
	{
		return val + 4*(*s - '0');
	}
	s--;
error:
	printf("Unknown register: %2.2s\n", s);
	longjmp(mainlp, 1); /* ++jrb */
}
#endif

PRIVATE long
value(s, s_p, seg_p)
char *s, **s_p;
int *seg_p;
{
	long k;
 
	if (*s == '\'') /* handle character constants here */
	{
		*s_p = s + 2;
		return s[1];
	}
	if (*s == '-' || isdigit(*s))
	{
		return strtol(s, s_p, 0);
	}
	if (*s == '$')
	{
		k = reg_addr(s+1);
		*s_p = s + 3;
		k = get_reg(curpid, k);

		/* XXX - check that ksize_to_physb is appropriate here.
		 * Also check there are no sign extension bugs from saddr and
		 * k being only long, while phys_bytes is now unsigned long.
		 */
		if (k >= saddr
		    && k < saddr + ksize_to_physb(prc->p_map[T].mem_len))
			*seg_p = T;
		else if (k >= daddr
			 && k < daddr + ksize_to_physb(prc->p_map[D].mem_len))
			*seg_p = D;
		else if (k >= taddr
			 && k < taddr + ksize_to_physb(prc->p_map[S].mem_len))
			*seg_p = D;
		return(k);
	}
	return lookup(s, s_p, seg_p);
}

#if (CHIP == INTEL)
extern struct nlist *findsname P_((char *name, int where, int tchar));

PRIVATE long
lookup(s, s_p, seg_p)
char *s, **s_p;
int *seg_p;
{
	char c;
	int l;
 
	struct nlist *symptr;

	for (l = 1; idchar(s[l]); ++l ) {}
	c = s[l]; s[l] = 0;
	
	if (strcmp("_start", s) == 0)
	{
		*seg_p = T;
		*(*s_p = s + 6) = c;
		return saddr;
	}
	if (strcmp("_data", s) == 0)
	{
		*seg_p = D;
		*(*s_p = s + 5) = c;
		return daddr;
	}
	if (strcmp("_stack", s) == 0)
	{
		*seg_p = S;
		*(*s_p = s + 6) = c;
		return taddr;
	}
	if ( (symptr = findsname(s, CSEG, FALSE )) != NULL)
	{
		*seg_p = T;
		*(*s_p = s + l) = c;
		return symptr->n_value + saddr;
	}
	if ( (symptr = findsname(s, DSEG, FALSE )) != NULL)
	{
		*seg_p = D;
		*(*s_p = s + l) = c;
		return symptr->n_value + saddr;
	}
	printf("%s: symbol not found\n", s);
	longjmp(mainlp, 1);
	return 0;		/* NOTREACHED */
}
#else
PRIVATE long
lookup(s, s_p, seg_p)
char *s, **s_p;
int *seg_p;
{
	extern struct nlist *nbufp;
	extern int nsyms;
	int i, j;
#ifdef GNU_SYMBOL
	char *tmps = malloc( strlen(s));
#define	FREETMPS free( tmps)
#else
	char tmps[ sizeof( nbufp[0].N_NAME)+1];
#define	FREETMPS 
#endif

	for (j = 0; idchar(s[j]); ++j )
	  tmps[ j] = s[ j];
	tmps[j] = '\0';
	
	if (strcmp("_start", tmps) == 0)
	{
		*seg_p = T;
		*s_p = s + 6;
		FREETMPS;
		return saddr;
	}
	if (strcmp("_data", tmps) == 0)
	{
		*seg_p = D;
		*s_p = s + 5;
		FREETMPS;
		return daddr;
	}
	if (strcmp("_stack", s) == 0)
	{
		*seg_p = S;
		*s_p = s + 6;
		FREETMPS;
		return taddr;
	}
	for (i = 0; i < nsyms; i++)
	        if (strcmp(nbufp[i].N_NAME, tmps)==0)
		{
		     switch (GET_CLASS (nbufp[i]))
			{
				case N_TEXT:
					*seg_p = T;
					break;
				case N_DATA:
				case N_BSS:
					*seg_p = D;
					break;
				default:
					continue;
			}
			*s_p = s + j;
		        FREETMPS;
			return nbufp[i].n_value + saddr;
		}

	printf("%s: symbol not found\n", tmps);
	FREETMPS;
	longjmp(mainlp, 1);
	return 0;		/* NOTREACHED */
}
#endif
