/* sym68.c for mdb */

/* This is not very machine-dependent and should never have been split from
 * sym86.c.  The main differences are that sym86.c gets a lot of utility
 * routines from db (they have been copied into this file), and this doesn't
 * support BSD a.out format and has weaker exec header and symbol table
 * handling routines.
 * 
 * This last statement is not true any more.
 */

#include "mdb.h"
#include <ctype.h>
#include <fcntl.h>
#include <unistd.h>

#if defined(__STDC__) || defined(__cplusplus)
# define P_(s) s
#else
# define P_(s) ()
#endif

extern int checkobj P_((struct EXEC_T *header));

FORWARD int compare P_((struct nlist *p1, struct nlist *p2));
FORWARD char *addr_to_name P_((long rel_addr, long *off_p));
FORWARD int read_header P_((int fd));
FORWARD void outsym P_((char *sp));
FORWARD int symprefix P_((char *t, struct nlist *sp));

PUBLIC int nsyms;
PUBLIC struct nlist *nbufp = (struct nlist *)NULL;
PRIVATE char *strp = (char *) NULL; 	/* buffer for string table */
PRIVATE struct EXEC_T header;		/* header of a.out file */

PRIVATE int compare(p1, p2)
struct nlist	*p1, *p2;
{
	if (p1->n_value < p2->n_value) return -1;
	if (p1->n_value > p2->n_value) return 1;
	return 0;
}

PUBLIC void getsyms(file)
char *file;
{
  int fd;
  struct nlist *np;

  fd = open(file, O_RDONLY);
  if (fd == -1) {
	fprintf(stderr, "can't open %s\n", file);
	return;
  }
  if (read_header(fd)) {
	fprintf(stderr, "%s: no executable file\n", file);
	close(fd);
	return;
  }
  if (header.a_syms != 0) {
    nbufp = (struct nlist *) malloc((size_t) (header.a_syms & 0xFFFF));
    if (nbufp == (struct nlist *)NULL) {
	fprintf(stderr, "db: out of memory on %s\n", file);
	exit(2);
	return;
    }
    if (read(fd, (char *) nbufp, (int) (header.a_syms & 0xFFFF))
        != (int) (header.a_syms & 0xFFFF)) {
	  fprintf(stderr, "%s: can't read symbol table\n", file);
	  free(nbufp);
	  close(fd);
	  return;
    }
    nsyms = (int) header.a_syms / sizeof(struct nlist);
    qsort(nbufp, nsyms, (int)sizeof(struct nlist), compare);
#ifdef GNU_SYMBOL
    if (header.a_strsize <= sizeof (header.a_strsize)) {
      fprintf(stderr, "%s: no string table\n", file);
      free(nbufp);
      close(fd);
      return;
    }
    strp = (char *) malloc((int) (header.a_strsize & 0xFFFF));
    if (strp == (char *)NULL) {
      fprintf(stderr, "db: out of memory on %s\n", file);
      exit(2);
      return;
    }
    if (read(fd, strp, (int) (header.a_strsize & 0xFFFF))
	!= (int) (header.a_strsize & 0xFFFF)) {
      fprintf(stderr, "%s: can't read string table\n", file);
      free(strp);
      free(nbufp);
      close(fd);
      return;
    }
    for (np = nbufp; np < nbufp + nsyms; np++)
      np->N_NAME = strp + np->n_un.n_strx;
    
#endif
    close(fd);
  }
}

PUBLIC void listsym(local_cmd)
char *local_cmd;
{
/* list all symbols matching an optional leading prefix and optional class */

    char *endprefix;
    char *prefix;
    char *sclass;
    int tchar;
    char zap, schar;
    int i;

    endprefix = prefix = skip(local_cmd + 1);
    if (isalpha((unsigned char) *endprefix) || *endprefix == '_')
	while (isalnum((unsigned char) *endprefix) || *endprefix == '_')
	    ++endprefix;
    sclass = skip(endprefix);
    if (*sclass == '\n' || *sclass == ';')
	tchar = '*';
    else
	tchar = *sclass;
    zap = *endprefix;
    *endprefix = 0;
    /* find and print all matching symbols */
    for (i = 0; i < nsyms; i++)
    {
      if (symprefix(prefix, &nbufp[i]))
	{
	  switch (GET_CLASS (nbufp[i]))
	    {
	    case N_UNDF:
		schar = 'u';
		if (nbufp[i].n_value != 0)
		    schar = 'c';
		break;
	    case N_ABS:
		schar = 'a';
		break;
	    case N_TEXT:
		schar = 't';
		break;
	    case N_DATA:
		schar = 'd';
		break;
	    case N_BSS:
		schar = 'b';
		break;
	    default:
		schar = '?';
		break;
	    }
	    if (GET_CLASS( nbufp[i]) == C_EXT &&
		schar != '?')
		schar += 'A' - 'a';
	    if (tchar != '*' && tchar != schar)
		continue;
	    if (schar == 'U')
		printf("        ");
	    else
		printf("%08lx", (schar == 'a' || schar == 'A')?
				nbufp[i].n_value
			:
				nbufp[i].n_value + saddr
			);
	    putchar(' ');
	    putchar(schar);
	    putchar(' ');
	    outsym(nbufp[i].N_NAME);
	    putchar('\n');
	}
    }	    
    *endprefix = zap;
}

PRIVATE char *addr_to_name(rel_addr, off_p)
long rel_addr, *off_p;
{
	register int i = 0, l = 0, r = nsyms;

	while (l < r)
	{
		i = (l + r) >> 1;
		if (rel_addr < nbufp[i].n_value) r = i;
		else if (rel_addr > nbufp[i].n_value) l = i + 1;
		else break;
	}

	if (l == nsyms || r == 0)
	{
		*off_p = rel_addr;
		return "_start";
	}
	while (i)
	{
		switch ( GET_CLASS(nbufp[i]))
		{
			case N_TEXT:
			case N_DATA:
			case N_BSS:
				break;

			default:
				i--;
				continue;
		}
		if (rel_addr < nbufp[i].n_value)
		{
			i--;
			continue;
		}
		break;
	}
	*off_p = rel_addr - nbufp[i].n_value;
	return nbufp[i].N_NAME;
}

PUBLIC void symbolic(addr, sep)
unsigned long addr;
int sep;
{
	long off;
	extern long saddr, eaddr;

	if (addr < saddr || addr > eaddr)
	{
		printf("0x%lx%c", addr, sep);
		return;
	}
	outsym(addr_to_name(addr - saddr, &off));
	if (off) printf("+0x%lx", off);
	fputc(sep, stdout);
}

PRIVATE int read_header(fd)
int fd;
{
  if (read(fd, (char *) &header, (int)sizeof(struct EXEC_T))
      != sizeof(struct EXEC_T))
    return(1);
  if (BADMAG(header)) return(1);
#ifdef GNU_SYMBOL
  header.a_syms -= header.a_strsize;
#endif
  lseek(fd, A_SYMPOS(header), SEEK_SET);
  if (!checkobj(&header))
  {
      close(fd);
      return(1);
  }

  return(0);
}

#ifndef GNU_SYMBOL
PRIVATE int symprefix(t, sp)
register char *t;
struct nlist *sp;
{
    register char *s;
    register char *send;

    for (; *t == '_'; ++t)
	;

    for (s = sp->N_NAME, send = s + sizeof sp->N_NAME;
	 s < send && *s == '_'; ++s)
	;

    for (; *t != 0 && s < send; ++s, ++t)
	if (*t != *s)
	    return FALSE;

    return TRUE;
}
#else
PRIVATE int symprefix(t, sp)
register char *t;
struct nlist *sp;
{
    register char *s;

    for (; *t == '_'; ++t)
	;

    for (s = sp->N_NAME; *s == '_'; ++s)
	;

    for (; *t != 0; ++s, ++t)
	if (*t != *s)
	    return FALSE;

    return TRUE;
}
#endif

#ifndef GNU_SYMBOL
PRIVATE void outsym(s)
register char *s;
{
    register char *send;

    for (send = s + sizeof nbufp->N_NAME; *s != 0 && s < send; ++s)
	putchar(*s);
}
#else
PRIVATE void outsym(s)
char *s;
{
   for (;*s != '\0'; s++)
	putchar(*s);
}
#endif
