/* sym86.c for mdb */

#include <sys/types.h>
#include <ctype.h>
#include <errno.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#ifdef BSD_A_OUT
/* Offsets of various things. This should be changed to use the Gnu access
 * macros.
 */
#define A_MINHDR	32
#define	A_TEXTPOS(X)	((long)(X).a_hdrlen)
#define A_DATAPOS(X)	(A_TEXTPOS(X) + (X).a_text)
#define	A_HASRELS(X)	((X).a_hdrlen > (unsigned char) A_MINHDR)
#define A_HASEXT(X)	((X).a_hdrlen > (unsigned char) (A_MINHDR +  8))
#define A_HASLNS(X)	((X).a_hdrlen > (unsigned char) (A_MINHDR + 16))
#define A_HASTOFF(X)	((X).a_hdrlen > (unsigned char) (A_MINHDR + 24))
#define A_TRELPOS(X)	(A_DATAPOS(X) + (X).a_data)
#define A_DRELPOS(X)	(A_TRELPOS(X) + (X).a_trsize)
#define A_SYMPOS(X)	(A_TRELPOS(X) + (A_HASRELS(X) ? \
				((X).a_trsize + (X).a_drsize) : 0))
#endif

#define FOR_MDB
#include "db/sym.c"		/* this does most of the work */

PRIVATE char *symfilename;

/* Prototypes created by mkptypes. */
#if defined(__STDC__) || defined(__cplusplus)
# define P_(s) s
#else
# define P_(s) ()
#endif

extern int checkobj P_((struct exec *header));
extern char *skip P((char *s));

FORWARD void symerr P_((char *msg));

PUBLIC void getsyms(filename)
char *filename;
{
    int fd;
    struct exec header;
    int nread;
#ifdef BSD_A_OUT    
    register struct nlist *sp;
    long strsize;
    char *strtable;
#endif    
    register struct symtab_s *tp;

    tp = &proc_sym[0];
    symfilename = filename;

    fd = open(filename, O_RDONLY);
    if (fd < 0)
    {
	symerr("1error opening");
	return;
    }
    nread = read(fd, (char *) &header, (unsigned) (sizeof header));
    if (nread < 0)
    {
	symerr("1error reading header from");
	close(fd);
	return;
    }
    if (nread != sizeof header)
    {
	symerr("0header too short in");
	close(fd);
	return;
    }
    if (!checkobj(&header))
    {
	close(fd);
	return;
    }
    if (lseek(fd, A_SYMPOS(header), SEEK_SET) != A_SYMPOS(header))
    {
	symerr("1error seeking to symbols in");
	close(fd);
	return;
    }
    if (header.a_syms != 0 && (header.a_syms != (int) header.a_syms ||
	(tp->start = malloc((size_t) header.a_syms)) == NULL))
    {
	symerr("0no room for symbol table for");
	close(fd);
	return;
    }
    nread = read(fd, (char *) tp->start, (unsigned) header.a_syms);
    if (nread < 0)
    {
	symerr("1error reading symbol table from");
	close(fd);
	return;
    }
    if (nread != (unsigned) header.a_syms)
    {
	symerr("0symbol table too short in");
	close(fd);
	return;
    }
#ifdef BSD_A_OUT
    nread = read(fd, (char *) &strsize, sizeof(long));
    if (nread != 0)
    {
	if (nread < 0)
	{
	    symerr("1error reading size of string table for");
	    close(fd);
	    return;
	}
	if (nread != sizeof(long))
	{
	    symerr("0string table header too short in");
	    close(fd);
	    return;
	}
	if (strsize < sizeof(long))
	{
	    symerr("0string table too short in");
	    close(fd);
	    return;
	}
	strsize -= sizeof(long);
	if (strsize != (int) strsize ||
	    (strtable = malloc((size_t) strsize)) == NULL)
	{
	    symerr("0no room for string table for");
	    close(fd);
	    return;
	}
	nread = read(fd, strtable, (unsigned) strsize);
	if (nread < 0)
	{
	    symerr("1error reading string table from");
	    close(fd);
	    return;
	}
	if (nread != strsize)
	{
	    symerr("0string table too short in");
	    close(fd);
	    return;
	}
    tp->nsym = (unsigned) header.a_syms / sizeof(struct nlist);
    tp->end = tp->start + tp->nsym;
	for (sp = tp->start; sp < tp->end; ++sp)
	    sp->n_un.n_name = strtable + sp->n_un.n_strx - 4;
    }
#endif /* BSD_A_OUT */
    close(fd);

    tp->stoff = 0;
    tp->nsym = (unsigned) header.a_syms / sizeof(struct nlist);
    tp->end = tp->start + tp->nsym;
    tp->text = 0;		/* segments are not known */

    /* sort on value only, name search not used much and storage a problem */
    outstr("Sorting ");
    outh16(tp->nsym);
    outspace();
    outstr(" symbols ... ");
    sortsyms(tp->start, tp->end, tp->stoff);
    outnl();
}

PUBLIC void listsym(local_cmd)
char *local_cmd;
{
/* list all symbols matching an optional leading prefix and optional class */

    char *endprefix;
    char *prefix;
    char *sclass;
    int tchar;
    char zap;

    endprefix = prefix = skip(local_cmd + 1);
    if (isalpha((unsigned char) *endprefix) || *endprefix == '_')
	while (isalnum((unsigned char) *endprefix) || *endprefix == '_')
	    ++endprefix;
    sclass = skip(endprefix);
    if (*sclass == '\n' || *sclass == ';')
	tchar = '*';
    else
	tchar = *sclass;
    zap = *endprefix;
    *endprefix = 0;
    findsname(prefix, CSEG, tchar);
    *endprefix = zap;
}

PUBLIC void symbolic(value, separator)
offset_T value;
int separator;
{
    register struct nlist *sp;

    if ((sp = findsval(value, CSEG)) != NULL)
    {
	outsym(sp, value);
    }
    outbyte(separator);
}

PRIVATE void symerr(msg)
char *msg;
{
    fprintf(stderr, "db: %s %s%s%s\n", msg + 1, symfilename,
	    *msg == '0' ? "" : ": ", *msg == '0' ? "" : strerror(errno));
}
