/*
 * Atari ST hardware:
 * DMA Sound Subsystem
 */

#define	DMA_SOUND	((struct sound *)AD_DMA_SOUND)

struct dma_sound {
	_VOLATILE unsigned char	snd[20];	/* use only the odd bytes */
	unsigned char gap[12];
	_VOLATILE unsigned short snd_mode_control;
	_VOLATILE unsigned short microwire_data;
	_VOLATILE unsigned short microwire_mask;
};

#define	snd_dma_control	snd[1]	/* select register */

/*
 * Indices into snd.
 * Access low byte of 16 bits.
 * Fill low/mid/high in this order.
 */

/* frame base address */
#define	SND_BASE_HIGH	3
#define	SND_BASE_MID	5
#define	SND_BASE_LOW	7

/* frame address counter */
#define	SND_CNT_HIGH	9
#define	SND_CNT_MID	11
#define	SND_CNT_LOW	13

/* frame end address */
#define	SND_END_HIGH	15
#define	SND_END_MID	17
#define	SND_END_LOW	19

/* bits in snd_dma_control */
#define ENABLE		0x01	/* 0 = disable, 1 = enable */
#define REPEAT		0x02	/* 0 = single frame, 1 = repeat */
/* other bits should be zero */

/* bits in snd_mode_control */
#define MODE		0x80	/* 0 = stereo, 1 = mono */
#define SAMPLE_RATE	0x03
/* other bits should be zero */

/* defines for SAMPLE RATE */
#define SF_6258		0	/* sample frequenc 6258 Hz */
#define SF_12517	1	/* sample frequenc 12517 Hz */
#define SF_25033	2	/* sample frequenc 25033 Hz */
#define SF_50066	3	/* sample frequenc 50066 Hz */

/* defines for MicroWire mask register */

#define LMC1992_MASK	0x07ff	/* mask for the LMC 1992 volume/tone control */

/* defines for MicroWire data register */

/* the following defines should be OR-ed with the wanted value */
#define LMC1992_MASTER_VOL	0x4c00	/* 0: -80 dB, 0x28: 0 dB, steps 2 dB */
#define LMC1992_LEFT_VOL	0x5400	/* 0: -40 dB, 0x14: 0 dB, steps 2 dB */
#define LMC1992_RIGHT_VOL	0x5000	/* 0: -40 dB, 0x14: 0 dB, steps 2 dB */
#define LMC1992_TREBLE		0x4800	/* 0: -12 dB, 0x6: 0 dB, 0xc: +12 dB */
#define LMC1992_BASS		0x4400	/* 0: -12 dB, 0x6: 0 dB, 0xc: +12 dB */
#define LMC1992_MIX		0x4000	/* 0: -12 dB, 1: mix with GI sound chip,
					   2: do not mix with GI sound chip */
