/*
 * Atari TT hardware:
 * MC146818A clock circuitry
 */

#define	RTC_NVM	((struct dma *)AD_RTC_NVM)

struct rtcnvm {
    unsigned char		rtc_gap1;
    _VOLATILE unsigned char	rtc_address;
    unsigned char		rtc_gap2;
    _VOLATILE unsigned char	rtc_data;
};

/*
 * Addresses
 */
#define	seconds		0x00
#define	seconds_alarm	0x01
#define	minuttes	0x02
#define	minutes_alarm	0x03
#define	hours		0x04
#define	hours_alarm	0x05
#define	weekday		0x06
#define	monthday	0x07
#define	month		0x08
#define	year		0x09
#define	register_a	0x0A
#define	register_b	0x0B
#define	register_c	0x0C
#define	register_d	0x0D

/* next 50 addresses are the non volatile memory */

/* register A bits */

#define UIP 		0x80	/* update in progress (read only) */
#define DV 		0x70	/* divider bits, should be 010 on TT */
#define RSEL 		0x0f	/* select bits, determine interrupt frequency
				   or square wave frequency */
/* register B bits */

#define SET 		0x80	/* if 0 update time 1/sec. if 1 no update */
#define PIE 		0x40	/* periodic interrupt enable */
#define AIE 		0x20	/* alarm interrupt enable */
#define UIE 		0x10	/* update ended interrupt enable */
#define SQWE 		0x08	/* square wave enable */
#define DM 		0x04	/* data mode (1 = binary mode, 0 = bcd mode) */
#define 24_12 		0x02	/* 12 or 24 hour mode (0 = 12 hour mode) */
#define DSE 		0x01	/* Daylight savings enable; 1 = set */
				/* Note: DSE updates for daylight saving time
				   	 on the last sunday in april and the 
				   	 last sunday in october, making this 
				   	 bit useless for use in Europe */
				   	 
/* register C bits */

#define IRQF 		0x80	/* 1 of any of PIE AIE UIE is 1, (read only) */
#define PF 		0x40	/* periodic interrupt flag, (read only) */
				/* PF is also set if PIE is off */
#define AF 		0x40	/* alarm interrupt flag, (read only) */
#define UF 		0x40	/* update-ended interrupt flag, (read only) */

/* register D bits */

#define VRT 		0x80	/* valid ram and time bit (read only) */
